#include <Wire.h>
#include <LiquidCrystal.h>


LiquidCrystal lcd(12, 11, 5, 4, 3, 2);


void i2c_write(int address, byte reg, byte data)
{
  // Send output register address
  Wire.beginTransmission(address);
  Wire.write(reg);
  // Connect to device and send byte
  Wire.write(data); // low byte
  Wire.endTransmission(true);    // stop
}


byte i2c_read(int address, byte reg)
{
  byte ret;
  // Send input register address
  Wire.beginTransmission(address);
  Wire.write(reg);
  Wire.endTransmission(false);	// restart
  Wire.requestFrom(address, 1);
  ret = Wire.read();
  Wire.endTransmission(true);  // stop
  return ret;
}

///////////////////////////////////////////////////////////////////////////
// PCA9698 pin assignment
//
// IO0_0-7  output  analog sensor enable (active H)
// IO1      input   NC
// IO2      input   NC
// IO3_0-7  output  LCD DB0-DB7
// IO4_0    output  LCD RS (H:data, L:command)
// IO4_1    output  LCD RW (H:read, W:write)
// IO4_2    output  LCD E (active H)
// IO4_3    output  LCD CS1 (active H)
// IO4_4    output  LCD CS2 (active H)
// IO4_5    output  LCD RST (active L)
// IO4_6-7  output  NC
///////////////////////////////////////////////////////////////////////////

#define  I2C_PCA9698_ADDR  (0x20 + 0)  // NOTE: 7bit address

byte    sns_select_g;    // global var; sensor select
byte    glcd_ctrl_g;     // global var; lcd control signals
int     sns_val_g[8];

void init_pca9698(int dev_addr)
{
  // set I/O mode
  i2c_write(dev_addr, 0x18, 0x00);     // IO0 output
  i2c_write(dev_addr, 0x19, 0xFF);     // IO1 input
  i2c_write(dev_addr, 0x1a, 0xFF);     // IO2 input
  i2c_write(dev_addr, 0x1b, 0x00);     // IO3 output
  i2c_write(dev_addr, 0x1c, 0x00);     // IO4 output

  // init output
  sns_select_g  = 0x01;
  glcd_ctrl_g   = 0x00;

  i2c_write(dev_addr, 0x08, sns_select_g);   // IO0
//  i2c_write(dev_addr, 0x09, 0x00);  // IO1
//  i2c_write(dev_addr, 0x0a, 0x00);  // IO2
  i2c_write(dev_addr, 0x0b, 0x00);    // IO3
  i2c_write(dev_addr, 0x0c, glcd_ctrl_g);    // IO4
}

/////////////////////////////////////////////////////////////////////////////////////////
// Graphic LCD driver

// level0: H/W pin dependent functions
void set_glcd_rst(byte val)
{
  if (val == 0)
    glcd_ctrl_g  &= 0xDF;
  else
    glcd_ctrl_g  |= 0x20;
  i2c_write(I2C_PCA9698_ADDR, 0x0c, glcd_ctrl_g);
}

void set_glcd_cs2(byte val)
{
  if (val == 0)
    glcd_ctrl_g  &= 0xEF;
  else
    glcd_ctrl_g  |= 0x10;
  i2c_write(I2C_PCA9698_ADDR, 0x0c, glcd_ctrl_g);
}

void set_glcd_cs1(byte val)
{
  if (val == 0)
    glcd_ctrl_g  &= 0xF7;
  else
    glcd_ctrl_g  |= 0x08;
  i2c_write(I2C_PCA9698_ADDR, 0x0c, glcd_ctrl_g);
}

void set_glcd_e(byte val)
{
  if (val == 0)
    glcd_ctrl_g  &= 0xFB;
  else
    glcd_ctrl_g  |= 0x04;
  i2c_write(I2C_PCA9698_ADDR, 0x0c, glcd_ctrl_g);
}

void set_glcd_rw(byte val)
{
  if (val == 0)
    glcd_ctrl_g  &= 0xFD;
  else
    glcd_ctrl_g  |= 0x02;
  i2c_write(I2C_PCA9698_ADDR, 0x0c, glcd_ctrl_g);
}

void set_glcd_rs(byte val)
{
  if (val == 0)
    glcd_ctrl_g  &= 0xFE;
  else
    glcd_ctrl_g  |= 0x01;
  i2c_write(I2C_PCA9698_ADDR, 0x0c, glcd_ctrl_g);
}

void glcd_send_cmd(byte data, byte mode, byte chip)
{
//  set_glcd_e(0);
  // (wait 3us)
  set_glcd_rs(mode);
//set_glcd_rw(0);
  i2c_write(I2C_PCA9698_ADDR, 0x0b, data);  // data bus
  if (chip == 0) {
    set_glcd_cs1(1);
    set_glcd_cs2(0);
  }
  else {
    set_glcd_cs1(0);
    set_glcd_cs2(1);
  }
  // (wait 3us)
  set_glcd_e(1);
  // (wait 3us)
  set_glcd_e(0);
}

// level 1: lcd comand (H/W independent)

void init_glcd(void)
{
  int x, y;

  set_glcd_rst(0);
  delay(50);
  set_glcd_rst(1);
  delay(50);
  set_glcd_rw(0);

  glcd_send_cmd(0xC0, 0, 0);    // CMD #0: set line
  glcd_send_cmd(0x3F, 0, 0);    // CMD #0: lcd on
  glcd_send_cmd(0xC0, 0, 1);    // CMD #1: set line
  glcd_send_cmd(0x3F, 0, 1);    // CMD #1: lcd on

  // clear memory
  for (x = 0; x < 8; x++) { 
    glcd_send_cmd(0xB8 + x, 0, 0);  // x addr
    glcd_send_cmd(0x40 + 0, 0, 0);  // y addr
    for (y = 0; y < 64; y++) {
      glcd_send_cmd(0x00, 1, 0);      // data
    }
  }
  for (x = 0; x < 8; x++) { 
    glcd_send_cmd(0xB8 + x, 0, 1);  // x addr
    glcd_send_cmd(0x40 + 0, 0, 1);  // y addr
    for (y = 0; y < 64; y++) {
      glcd_send_cmd(0x00, 1, 1);      // data
    }
  }

//    glcd_send_cmd(0xB8 + 1, 0, 0);  // x addr
//    glcd_send_cmd(0x40 + 1, 0, 0);  // y addr
//    glcd_send_cmd(0x5A, 1, 0);      // data
}

void draw_glcd(byte xpos, byte ypos, byte data)
{
  if (ypos < 64) {
    glcd_send_cmd(0xB8 + xpos, 0, 0);  // x addr
    glcd_send_cmd(0x40 + ypos, 0, 0);  // y addr
    glcd_send_cmd(data, 1, 0);         // data
  }
  else {
    glcd_send_cmd(0xB8 + xpos, 0, 1);  // x addr
    glcd_send_cmd(0x40 + (ypos - 64), 0, 1);  // y addr
    glcd_send_cmd(data, 1, 1);         // data
  }
}

// level 2: application dependent
void draw_bar_glcd(byte sns, int data)  // data:0-1023
{
  static int bar_length[8]  = {-1, -1, -1, -1, -1, -1, -1, -1};
  int  pos;

  data &= 0x3FF;
  data >>= 2;

  if (data == bar_length[sns])
    return;
  bar_length[sns]  = data;

  for (pos = 0; pos < 128; pos++) {
    if (pos <= data)
      draw_glcd(sns, pos, 0x7E);
    else
      draw_glcd(sns, pos, 0x00);
  }
}


/////////////////////////////////////////////////////////////////////////////////////////

void setup()
{
  // I2C
  Wire.begin(); // join i2c bus (address optional for master)
  init_pca9698(I2C_PCA9698_ADDR);

  // charactor LCD
  lcd.begin(16, 2);
  lcd.clear();
  lcd.display();

  // graphic LCD
  init_glcd();
}

void loop()
{
  int  v;
  char disp_buf[32];

  i2c_write(I2C_PCA9698_ADDR, 0x08, sns_select_g);  // set analog sw
  delayMicroseconds(200);
  v  = analogRead(0);                       // read port

  switch (sns_select_g) {
  case 0x01:
    sns_val_g[0]  = v;
    sns_select_g  = 0x02;
    draw_bar_glcd(0, v);
    break;
  case 0x02:
    sns_val_g[1]  = v;
    sns_select_g  = 0x04;
    draw_bar_glcd(1, v);
    break;
  case 0x04:
    sns_val_g[2]  = v;
    sns_select_g  = 0x08;
    draw_bar_glcd(2, v);
    break;
  case 0x08:
    sns_val_g[3]  = v;
    sns_select_g  = 0x10;
    draw_bar_glcd(3, v);
    break;
  case 0x10:
    sns_val_g[4]  = v;
    sns_select_g  = 0x20;
    draw_bar_glcd(4, v);
    break;
  case 0x20:
    sns_val_g[5]  = v;
    sns_select_g  = 0x40;
    draw_bar_glcd(5, v);
    break;
  case 0x40:
    sns_val_g[6]  = v;
    sns_select_g  = 0x80;
    draw_bar_glcd(6, v);
    break;
  case 0x80:
    sns_val_g[7]  = v;
    sns_select_g  = 0x01;
    draw_bar_glcd(7, v);

    lcd.setCursor(0, 0);
    sprintf(disp_buf, "%d,%d,%d,%d    ", sns_val_g[0], sns_val_g[1], sns_val_g[2], sns_val_g[3]);
    lcd.print(disp_buf);

    lcd.setCursor(0, 1);
    sprintf(disp_buf, "%d,%d,%d,%d    ", sns_val_g[4], sns_val_g[5], sns_val_g[6], sns_val_g[7]);
    lcd.print(disp_buf);

    break;

  default:
    sns_select_g  = 0x01;
  }
}
