///////////////////////////////////////////////////////
// I2C Slave core
// 2012.11.25
///////////////////////////////////////////////////////

module  i2c_slave_core(
			i2c_sda_in, i2c_sda_out, i2c_sda_oe, i2c_scl_in,
			dev_addr, reg_addr_out, reg_data_in, reg_data_out, reg_read_en_out, reg_write_en_out,
			rst_n, clk
		);

	///////////////////////////////////////////////////////
	// PI, PO
	///////////////////////////////////////////////////////

	// i2c side
	input			i2c_sda_in;			// UNsynchronized external signal (PI)
	output			i2c_sda_out;		// data output
	output			i2c_sda_oe;			// data output enable (for controlling bi-directional port)

	input			i2c_scl_in;			// UNsynchronized external signal (PI)

	// core side
	input [6:0]		dev_addr;

	output [7:0]	reg_addr_out;
	input [7:0]		reg_data_in;
	output [7:0]	reg_data_out;
	output			reg_read_en_out;	// data read strobe
	output			reg_write_en_out;	// data write strobe

	// clk, reset
	input			rst_n;				// sync
	input			clk;

	///////////////////////////////////////////////////////
	// registers
	///////////////////////////////////////////////////////
	reg			i2c_sda_out_reg;		// PO
	reg			i2c_sda_oe_reg;			// PO

	reg [7:0]	reg_addr_reg;			// PO
	reg [7:0]	reg_data_reg;			// PO
	reg			reg_write_reg;			// PO
	reg			reg_read_reg;			// PO

	reg			pre_i2c_sda_reg;
	reg [6:0]	dev_addr_reg;
	reg			ignore_reg;
	reg [3:0]	bit_cnt_reg;

	parameter [5:0]	ST_H0		= 6'b000000;
	parameter [5:0]	ST_ST0		= 6'b000001;
	parameter [5:0]	ST_L0		= 6'b000010;
	parameter [5:0]	ST_DEVA0	= 6'b010000;
	parameter [5:0]	ST_DEVA1	= 6'b010001;
	parameter [5:0]	ST_DEVA2	= 6'b010010;
	parameter [5:0]	ST_DEVA3	= 6'b010011;
	parameter [5:0]	ST_DEVA4	= 6'b010100;
	parameter [5:0]	ST_DEVA5	= 6'b010101;
	parameter [5:0]	ST_REGA0	= 6'b010110;
	parameter [5:0]	ST_REGA1	= 6'b010111;
	parameter [5:0]	ST_REGA2	= 6'b011000;
	parameter [5:0]	ST_REGA3	= 6'b011001;
	parameter [5:0]	ST_DET0		= 6'b011010;
	parameter [5:0]	ST_DET1		= 6'b011011;
	parameter [5:0]	ST_DET2		= 6'b011100;
	parameter [5:0]	ST_WR0		= 6'b100000;
	parameter [5:0]	ST_WR1		= 6'b100001;
	parameter [5:0]	ST_WR2		= 6'b100010;
	parameter [5:0]	ST_WR3		= 6'b100011;
	parameter [5:0]	ST_WR4		= 6'b100100;
	parameter [5:0]	ST_RD0		= 6'b110000;
	parameter [5:0]	ST_RD1		= 6'b110001;
	parameter [5:0]	ST_RD2		= 6'b110010;
	parameter [5:0]	ST_RD3		= 6'b110011;
	parameter [5:0]	ST_RD4		= 6'b110100;
	parameter [5:0]	ST_RD5		= 6'b110101;
	parameter [5:0]	ST_RD6		= 6'b110110;
	parameter [5:0]	ST_RD7		= 6'b110111;
	parameter [5:0]	ST_RD8		= 6'b111000;
	parameter [5:0]	ST_RD9		= 6'b111001;
	reg [5:0]	state_reg;

	///////////////////////////////////////////////////////
	// synchronizers
	///////////////////////////////////////////////////////
	reg	[3:0]	i2c_sda_sync_reg;
	reg	[3:0]	i2c_scl_sync_reg;

	wire		i2c_sda_in_sync;
	wire		i2c_scl_in_sync;

	always @(posedge clk) begin		// synchronizer
		if (rst_n == 1'b0) begin
			i2c_sda_sync_reg	<= 4'hF;
			i2c_scl_sync_reg	<= 4'hF;
		end else begin
			i2c_scl_sync_reg[2:0]	<= {i2c_scl_sync_reg[1:0], i2c_scl_in};

			if (i2c_scl_sync_reg[2:1] == 2'b11)
				i2c_scl_sync_reg[3]	<= 1'b1;
			else if (i2c_scl_sync_reg[2:1] == 2'b00)
				i2c_scl_sync_reg[3]	<= 1'b0;

			i2c_sda_sync_reg[2:0]	<= {i2c_sda_sync_reg[1:0], i2c_sda_in};

			if (i2c_sda_sync_reg[2:1] == 2'b11)
				i2c_sda_sync_reg[3]	<= 1'b1;
			else if (i2c_sda_sync_reg[2:1] == 2'b00)
				i2c_sda_sync_reg[3]	<= 1'b0;

		end
	end

	assign	i2c_sda_in_sync	= i2c_sda_sync_reg[3];
	assign	i2c_scl_in_sync	= i2c_scl_sync_reg[3];


	///////////////////////////////////////////////////////
	// main control
	///////////////////////////////////////////////////////
	always @(posedge clk) begin
		if (rst_n == 1'b0) begin
			i2c_sda_out_reg	<= 1'b1;
			i2c_sda_oe_reg	<= 1'b0;

			reg_addr_reg	<= 8'd0;
			reg_data_reg	<= 8'd0;
			reg_write_reg	<= 1'b0;
			reg_read_reg	<= 1'b0;

			state_reg	<= ST_H0;

		end else begin
			pre_i2c_sda_reg	<= i2c_sda_in_sync;

			case (state_reg)
			//////////////////////////////////////////////////
			// wait start
			//////////////////////////////////////////////////
			ST_H0: begin		// wait until scl == 1 && sda == 1
				if (i2c_scl_in_sync == 1'b1 && i2c_sda_in_sync == 1'b1) begin
					state_reg		<= ST_ST0;
				end
			end

			ST_ST0: begin		// wait START
				if (i2c_scl_in_sync == 1'b0) begin
					state_reg	<= ST_H0;
				end else if (i2c_sda_in_sync == 1'b0 && pre_i2c_sda_reg == 1'b1) begin
					state_reg	<= ST_L0;
				end else if (i2c_sda_in_sync == 1'b1 && pre_i2c_sda_reg == 1'b0) begin	// STOP
					i2c_sda_oe_reg	<= 1'b0;
					state_reg		<= ST_H0;
				end
			end

			ST_L0: begin		// wait until scl == 0
				if (i2c_scl_in_sync == 1'b0) begin
					bit_cnt_reg	<= 4'd0;
					state_reg	<= ST_DEVA0;
				end else if (i2c_sda_in_sync == 1'b1 && pre_i2c_sda_reg == 1'b0) begin	// STOP
					i2c_sda_oe_reg	<= 1'b0;
					state_reg		<= ST_H0;
				end
			end

			//////////////////////////////////////////////////
			// receive device address
			//////////////////////////////////////////////////
			ST_DEVA0: begin	// wait until scl == 1
				if (i2c_scl_in_sync == 1'b1) begin
					dev_addr_reg	<= {dev_addr_reg[5:0], i2c_sda_in_sync};
					bit_cnt_reg		<= bit_cnt_reg + 1'b1;
					state_reg		<= ST_DEVA1;
				end
			end

			ST_DEVA1: begin	// loop
				if (i2c_scl_in_sync == 1'b0) begin
					if (bit_cnt_reg >= 4'd7) begin
						state_reg	<= ST_DEVA2;
						if (dev_addr_reg == dev_addr) begin
							ignore_reg	<= 1'b0;
						end else begin
							ignore_reg	<= 1'b1;
						end
					end else begin
						state_reg	<= ST_DEVA0;
					end
				end else if (i2c_sda_in_sync == 1'b1 && pre_i2c_sda_reg == 1'b0) begin	// STOP
					i2c_sda_oe_reg	<= 1'b0;
					state_reg		<= ST_H0;
				end
			end

			ST_DEVA2: begin	// wait until scl == 1
				if (i2c_scl_in_sync == 1'b1) begin
					if (i2c_sda_in_sync == 1'b1) begin	// not write
						ignore_reg	<= 1'b1;
					end
					state_reg	<= ST_DEVA3;
				end
			end

			ST_DEVA3: begin	// wait until scl == 0
				if (i2c_scl_in_sync == 1'b0) begin
					i2c_sda_out_reg	<= ignore_reg;		// ACK:0, NACK:1
					i2c_sda_oe_reg	<= ~ignore_reg;		// ACK:1, NACK:0
					state_reg		<= ST_DEVA4;
				end else if (i2c_sda_in_sync == 1'b1 && pre_i2c_sda_reg == 1'b0) begin	// STOP
					i2c_sda_oe_reg	<= 1'b0;
					state_reg		<= ST_H0;
				end
			end

			ST_DEVA4: begin	// wait until scl == 1 (return ACK/NACK)
				if (i2c_scl_in_sync == 1'b1) begin
					state_reg	<= ST_DEVA5;
				end
			end

			ST_DEVA5: begin	// wait until scl == 0
				if (i2c_scl_in_sync == 1'b0) begin
					i2c_sda_oe_reg	<= 1'b0;
					bit_cnt_reg		<= 4'd0;
					state_reg		<= ST_REGA0;
				end
			end

			//////////////////////////////////////////////////
			// receive register address
			//////////////////////////////////////////////////
			ST_REGA0: begin	// wait until scl == 1
				if (i2c_scl_in_sync == 1'b1) begin
					reg_addr_reg	<= {reg_addr_reg[6:0], i2c_sda_in_sync};
					bit_cnt_reg		<= bit_cnt_reg + 1'b1;
					state_reg		<= ST_REGA1;
				end
			end

			ST_REGA1: begin	// loop
				if (i2c_scl_in_sync == 1'b0) begin
					if (bit_cnt_reg >= 4'd8) begin
						i2c_sda_out_reg	<= ignore_reg;		// ACK:0, NACK:1
						i2c_sda_oe_reg	<= ~ignore_reg;		// ACK:1, NACK:0
						state_reg		<= ST_REGA2;
					end else begin
						state_reg	<= ST_REGA0;
					end
				end else if (i2c_sda_in_sync == 1'b1 && pre_i2c_sda_reg == 1'b0) begin	// STOP
					i2c_sda_oe_reg	<= 1'b0;
					state_reg		<= ST_H0;
				end
			end

			ST_REGA2: begin	// wait until scl == 1 (return ACK/NACK)
				if (i2c_scl_in_sync == 1'b1) begin
					state_reg	<= ST_REGA3;
				end
			end

			ST_REGA3: begin	// wait until scl == 0
				if (i2c_scl_in_sync == 1'b0) begin
					i2c_sda_oe_reg	<= 1'b0;
					state_reg		<= ST_DET0;
				end
			end

			//////////////////////////////////////////////////
			// determine read or write
			//////////////////////////////////////////////////
			ST_DET0: begin	// wait until scl == 1
				if (i2c_scl_in_sync == 1'b1) begin
					bit_cnt_reg		<= 4'd1;
					reg_data_reg	<= {reg_data_reg[6:0], i2c_sda_in_sync};
					state_reg		<= ST_DET1;
				end
			end

			ST_DET1: begin	// wait until START or scl == 0
				if (i2c_scl_in_sync == 1'b0) begin	// SCL == 0 (write transaction)
					state_reg	<= ST_WR0;
				end else if (i2c_sda_in_sync == 1'b0 && pre_i2c_sda_reg == 1'b1) begin	// START (read transaction)
					state_reg	<= ST_DET2;
				end else if (i2c_sda_in_sync == 1'b1 && pre_i2c_sda_reg == 1'b0) begin	// STOP
					i2c_sda_oe_reg	<= 1'b0;
					state_reg		<= ST_H0;
				end
			end

			ST_DET2: begin
				if (i2c_scl_in_sync == 1'b0) begin
					bit_cnt_reg	<= 4'd0;
					state_reg	<= ST_RD0;
				end else if (i2c_sda_in_sync == 1'b1 && pre_i2c_sda_reg == 1'b0) begin	// STOP
					i2c_sda_oe_reg	<= 1'b0;
					state_reg		<= ST_H0;
				end
			end

			//////////////////////////////////////////////////
			// register write transaction
			//////////////////////////////////////////////////
			ST_WR0: begin	// wait until scl == 1
				if (i2c_scl_in_sync == 1'b1) begin
					bit_cnt_reg		<= bit_cnt_reg + 1'b1;
					reg_data_reg	<= {reg_data_reg[6:0], i2c_sda_in_sync};
					state_reg		<= ST_WR1;
				end
			end

			ST_WR1: begin	// wait unitl scl == 0
				if (i2c_scl_in_sync == 1'b0) begin
					if (bit_cnt_reg >= 4'd8) begin
						i2c_sda_out_reg	<= ignore_reg;		// ACK:0, NACK:1
						i2c_sda_oe_reg	<= ~ignore_reg;		// ACK:1, NACK:0
						state_reg		<= ST_WR2;
					end else begin
						state_reg		<= ST_WR0;
					end
				end else if (i2c_sda_in_sync == 1'b1 && pre_i2c_sda_reg == 1'b0) begin	// STOP
					i2c_sda_oe_reg	<= 1'b0;
					state_reg		<= ST_H0;
				end
			end

			ST_WR2: begin	// wait until scl == 1 (return ACK/NACK)
				if (i2c_scl_in_sync == 1'b1) begin
					state_reg	<= ST_WR3;
				end
			end

			ST_WR3: begin	// wait until scl == 0
				if (i2c_scl_in_sync == 1'b0) begin
					i2c_sda_oe_reg	<= 1'b0;
					reg_write_reg	<= ~ignore_reg;
					state_reg		<= ST_WR4;
				end
			end

			ST_WR4: begin
				reg_write_reg	<= 1'b0;
				bit_cnt_reg		<= 4'd0;
				state_reg		<= ST_WR0;
			end

			//////////////////////////////////////////////////
			// register read transaction
			//////////////////////////////////////////////////
			ST_RD0: begin	// wait until scl == 1
				if (i2c_scl_in_sync == 1'b1) begin
					dev_addr_reg	<= {dev_addr_reg[5:0], i2c_sda_in_sync};
					bit_cnt_reg		<= bit_cnt_reg + 1'b1;
					state_reg		<= ST_RD1;
				end
			end

			ST_RD1: begin	// loop
				if (i2c_scl_in_sync == 1'b0) begin
					if (bit_cnt_reg >= 4'd7) begin
						state_reg	<= ST_RD2;
						if (dev_addr_reg == dev_addr) begin
							ignore_reg	<= 1'b0;
						end else begin
							ignore_reg	<= 1'b1;
						end
					end else begin
						state_reg	<= ST_RD0;
					end
				end else if (i2c_sda_in_sync == 1'b1 && pre_i2c_sda_reg == 1'b0) begin	// STOP
					i2c_sda_oe_reg	<= 1'b0;
					state_reg		<= ST_H0;
				end
			end

			ST_RD2: begin	// wait until scl == 1
				if (i2c_scl_in_sync == 1'b1) begin
					if (i2c_sda_in_sync == 1'b0) begin	// not read
						ignore_reg	<= 1'b1;
					end
					state_reg	<= ST_RD3;
				end
			end

			ST_RD3: begin	// wait until scl == 0
				if (i2c_scl_in_sync == 1'b0) begin
					i2c_sda_out_reg	<= ignore_reg;		// ACK:0, NACK:1
					i2c_sda_oe_reg	<= ~ignore_reg;		// ACK:1, NACK:0
					reg_read_reg	<= ~ignore_reg;
					state_reg		<= ST_RD4;
				end else if (i2c_sda_in_sync == 1'b1 && pre_i2c_sda_reg == 1'b0) begin	// STOP
					i2c_sda_oe_reg	<= 1'b0;
					state_reg		<= ST_H0;
				end
			end

			ST_RD4: begin	// wait until scl == 1 (return ACK/NACK)
				if (reg_read_reg == 1'b1) begin
					reg_data_reg	<= reg_data_in;
					reg_read_reg	<= 1'b0;
				end

				if (i2c_scl_in_sync == 1'b1) begin
//reg_data_reg	<= reg_addr_reg + 1'b1;
					state_reg		<= ST_RD5;
				end
			end

			ST_RD5: begin	// wait until scl == 0
				if (i2c_scl_in_sync == 1'b0) begin
					bit_cnt_reg		<= 4'd0;
					i2c_sda_out_reg	<= reg_data_reg[7];
					i2c_sda_oe_reg	<= ~ignore_reg;
					state_reg		<= ST_RD6;
				end
			end

			// output data
			ST_RD6: begin	// wait until scl == 1
				if (i2c_scl_in_sync == 1'b1) begin
					bit_cnt_reg		<= bit_cnt_reg + 1'b1;
					reg_data_reg	<= {reg_data_reg[6:0], 1'b1};
					state_reg		<= ST_RD7;
				end
			end

			ST_RD7: begin	// wait until scl == 0 or STOP
				if (i2c_scl_in_sync == 1'b0) begin
					if (bit_cnt_reg >= 4'd8) begin
						i2c_sda_oe_reg	<= 1'b0;
						state_reg		<= ST_RD8;
					end else begin
						i2c_sda_out_reg	<= reg_data_reg[7];
						state_reg		<= ST_RD6;
					end
				end else if (i2c_sda_in_sync == 1'b1 && pre_i2c_sda_reg == 1'b0) begin	// STOP
					i2c_sda_oe_reg	<= 1'b0;
					state_reg		<= ST_H0;
				end
			end

			// wait ack
			ST_RD8: begin	// wait until scl == 1
				if (i2c_scl_in_sync == 1'b1) begin
					ignore_reg	<= i2c_scl_in_sync;		// if NACK is returned, set ignore flag
					state_reg	<= ST_RD5;
				end
			end


			default: begin
				i2c_sda_out_reg	<= 1'b1;
				i2c_sda_oe_reg	<= 1'b0;
				reg_addr_reg	<= 8'd0;
				reg_data_reg	<= 8'd0;
				reg_read_reg	<= 1'b0;
				reg_write_reg	<= 1'b0;
				state_reg		<= ST_H0;
			end
			endcase
		end
	end

	///////////////////////////////////////////////////////
	// PO
	///////////////////////////////////////////////////////
	assign	i2c_sda_out			= i2c_sda_out_reg;
	assign	i2c_sda_oe			= i2c_sda_oe_reg;

	assign	reg_addr_out		= reg_addr_reg;
	assign	reg_data_out		= reg_data_reg;
	assign	reg_write_en_out	= reg_write_reg;
	assign	reg_read_en_out		= reg_read_reg;

endmodule
// end of file
